﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

namespace ZoomLcdChar
{
    public partial class zoomLcdChar : UserControl
    {
        private LanguageZoomLcdChar _lang; // Здесь хранятся все строки на нужном языке.
        private readonly bool[,] _clip; // Имитация буфера обмена.
        private Size _addrSize; // Размер строки адреса.
        public zoomLcdChar()
        {
            InitializeComponent();
            // Установка языка по-умолчанию (русский).
            SetLanguage(new LanguageZoomLcdChar());
            // Для лучшего отображения (кстранение мерцания при перерисовке и т.д. 
            SetStyle(ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint | ControlStyles.OptimizedDoubleBuffer | ControlStyles.ResizeRedraw, true);
            // Инициализация массивов точек и буфера обмена.
            _dots = new bool[5, 8];
            _clip = new bool[5, 8];
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8; y++)
                    _dots[x, y] = _clip[x, y] = false;
        }
        /// <summary>
        /// Установка языка (всплывающие подсказки и надпись "адрес").
        /// </summary>
        /// <param name="lang">Класс в файле Language.cs</param>
        public void SetLanguage(LanguageZoomLcdChar lang)
        {
            _lang = lang;
            // Измерение размеров строки адреса символа.
            _addrSize = TextRenderer.MeasureText(_lang.Address, Font, new Size(int.MaxValue, int.MaxValue));
            // Установка всплывающих подсказок.
            tTip.SetToolTip(charClear, _lang.Clear);
            tTip.SetToolTip(charFill, _lang.Fill);
            tTip.SetToolTip(charInverse, _lang.Inverse);
            tTip.SetToolTip(charLeft, _lang.SdvigLeft);
            tTip.SetToolTip(charUp, _lang.SdvigUp);
            tTip.SetToolTip(charDown, _lang.SdvigDown);
            tTip.SetToolTip(charRight, _lang.SdvigRight);
            tTip.SetToolTip(charLeftRight, _lang.FlipHoriz);
            tTip.SetToolTip(charUpDown, _lang.FlipVert);
            tTip.SetToolTip(charCopy, _lang.Copy);
            tTip.SetToolTip(charCut, _lang.Cut);
            tTip.SetToolTip(charPaste, _lang.Paste);
        }

        /// <summary>
        /// Рисование, т.е. обработка события onPaint.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ZoomLcdCharPaint(object sender, PaintEventArgs e)
        {
            SolidBrush b = new SolidBrush(ForeColor);
            Pen p = new Pen(_dotColor);

            e.Graphics.Clear(_fonColor); // Очистка области рисования, закрашиванием ее цветом фона.
            // Вывод строки адреса.
            e.Graphics.DrawString(_lang.Address, Font, b, Margin.Left, Margin.Top);
            b = new SolidBrush(_addrColor);
            e.Graphics.DrawString(" 0x" + _address.X.ToString("X") + _address.Y.ToString("X"), Font, b, Margin.Left + _addrSize.Width, Margin.Top);
            // Вывод точек матрицы и их границ.
            b = new SolidBrush(_dotColor);
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8; y++)
                {
                    if (_dots[x, y]) e.Graphics.FillRectangle(b, 5 + 16 * x, _addrSize.Height + Margin.Top * 2 + 16 * y, 15, 15);
                    else e.Graphics.DrawRectangle(p, 5 + 16 * x, _addrSize.Height + Margin.Top * 2 + 16 * y, 14, 14);
                }
            /* Уничтожение кисти и карандаша. 
             * На самом деле, сборщик мусора и сам бы уничтожил, но лучше всё-таки
             * это делать самостоятельно, особенно, если неизвестно, уничтожит или нет, 
             * есть он в этой среде или нет...*/
            b.Dispose();
            p.Dispose();
        }

        /// <summary>
        /// Обработка события onMouseDown.
        /// </summary>
        private void ZoomLcdCharMouseDown(object sender, MouseEventArgs e)
        {
            // Определение начальной высоты поля, где отрисованы "точки" символа.
            int tt = _addrSize.Height + Margin.Top * 2;
            // Если запрещено редактирование или кнопка мыши не левая или курсор за пределами поля - выход из функции.
            if (!_canEdit || (e.Button != MouseButtons.Left) ||
               (e.X < 5) || (e.X > Width - 6) || (e.Y < tt) || (e.Y > tt + 126)) return;
            // Определение, в какую точку попал курсор.
            byte y = (byte)((e.Y - tt) / 16);
            byte x = (byte)((e.X - 5) / 16);
            // Если получились недопустимые значения - приведем к допустимым. 
            if (y > 8) y = 8;
            if (x > 5) x = 5;
            if (y < 0) y = 0;
            if (x < 0) x = 0;
            // Переключение точки в противоположное состояние.
            _dots[x, y] = !_dots[x, y];
            // Вызываем функцию, где вызывается перерисовка и генерация события.
            updatevent();
        } 

        #region toolbar
        #region transform
        /// <summary>
        /// Здесь простые обработки некоторого инструментария с точками.
        /// </summary>
        private void FillClearDots(bool a)  // Закрашивание или очистка всех точек.
        {
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8; y++)
                    _dots[x, y] = a;
            updatevent();
        }
        private void InversDots() // Инверсия точек.
        {
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8; y++)
                    _dots[x, y] = !_dots[x, y];
            updatevent();
        }
        private void SdvigUp(bool a) // Сдвиг точек вверх. Здесь и далее операнд "а" определяет, что делать с "новыми" точками, 
        {                            // появившимися после сдвига, закрвшивать их, или оставлять пустыми.
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8 - 1; y++)
                    _dots[x, y] = _dots[x, y + 1];
            for (byte x = 0; x < 5; x++)
                _dots[x, 8 - 1] = a;
            updatevent();
        }
        private void SdvigDown(bool a) // Сдвиг точек вниз
        {
            for (byte x = 0; x < 5; x++)
                for (byte y = 8 - 1; y > 0; y--)
                    _dots[x, y] = _dots[x, y - 1];
            for (byte x = 0; x < 5; x++)
                _dots[x, 0] = a;
            updatevent();
        }
        private void SdvigRight(bool a) // Сдвиг точек вправо
        {
            for (byte y = 0; y < 8; y++)
            {
                for (byte x = 5 - 1; x > 0; x--) _dots[x, y] = _dots[x - 1, y];
                _dots[0, y] = a;
            }
            updatevent();
        }
        private void SdvigLeft(bool a) // Сдвиг точек влево
        {
            for (byte y = 0; y < 8; y++)
            {
                for (byte x = 0; x < 4; x++) _dots[x, y] = _dots[x + 1, y];
                _dots[4, y] = a;
            }
            updatevent();
        }
        private void FlipHoriz() // Отражение по горизонтали
        {
            bool[] a = new bool[5];
            for (byte y = 0; y < 8; y++)
            {
                byte z = 4;
                for (byte x = 0; x < 5; x++) a[z--] = _dots[x, y];
                for (byte x = 0; x < 5; x++) _dots[x, y] = a[x];
            }
            updatevent();
        }
        private void FlipVert() // отражение по вертикали
        {
            bool[] a = new bool[8];
            for (byte x = 0; x < 5; x++)
            {
                byte z = 7;
                for (byte y = 0; y < 8; y++) a[z--] = _dots[x, y];
                for (byte y = 0; y < 8; y++) _dots[x, y] = a[y];
            }
            updatevent();
        }
        #endregion
        #region cut-copy-paste

        /* Работа с "буфером обмена". Это будет необходимо в дальнейшем, для работы в комплексе симулятора
         * "Буфер обмена" не случайно взят в кавычки - он не является буфером обмена ОС. Впрочем, если нужно,
         * то это совершенно несложно добавить.
         */
        private void Paste() // Вставка из буфера.
        {
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8; y++)
                    _dots[x, y] = _clip[x, y];
            updatevent();
        }
        private void Cut(bool a) // Вырез в буфер.
        {
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8; y++)
                {
                    _clip[x, y] = _dots[x, y];
                    _dots[x, y] = a;
                }
            updatevent();
        }
        private void Copy() // Копирование в буфер.
        {
            for (byte x = 0; x < 5; x++)
                for (byte y = 0; y < 8; y++)
                    _clip[x, y] = _dots[x, y];
        }
        #endregion
        /// <summary>
        /// Обработка события onMouseClick на кнопках
        /// </summary>
        private void charTools_Click(object sender, MouseEventArgs e)
        {    
            if (((Button)sender).Name == "charCopy") // копирование разрешено всегда
                Copy();
            else if (_canEdit) // если разрешено редактирование, то...
                switch (((Button)sender).Name)
                {
                    case "charClear":
                        FillClearDots(false);
                        break;
                    case "charFill":
                        FillClearDots(true);
                        break;
                    case "charInverse":
                        InversDots();
                        break;
                    case "charUp": // Здесь и далее определяется, что делать с "новыми" точками после сдвига (закрасить или очистить),
                        SdvigUp(e.Button != MouseButtons.Left); // в зависимости от того, какая кнопка мыши была нажата.
                        break;
                    case "charLeft":
                        SdvigLeft(e.Button != MouseButtons.Left);
                        break;
                    case "charRight":
                        SdvigRight(e.Button != MouseButtons.Left);
                        break;
                    case "charDown":
                        SdvigDown(e.Button != MouseButtons.Left);
                        break;
                    case "charUpDown":
                        FlipVert();
                        break;
                    case "charLeftRight":
                        FlipHoriz();
                        break;
                    case "charCut":
                        Cut(e.Button != MouseButtons.Left);
                        break;
                    case "charPaste":
                        Paste();
                        break;
                }
        }
        #endregion
        #region event
        private void updatevent()
        {
            Invalidate(); // Вызов перерисовки.
            OnZoomLcdCharChanged(new EventArgs()); // Генерация события, что символ изменен.
        }
        public event EventHandler<EventArgs> ZoomLcdCharChanged;
        protected virtual void OnZoomLcdCharChanged(EventArgs e)
        {
            ZoomLcdCharChanged?.Invoke(this, e);
        }
        #endregion
        #region property
        /// <summary>
        /// Поскольку VisualStudio не может явно сохранять в свойствах компонентов двумерные массивы,
        /// доступ к матрице символа сделан через фукции GetDos и SetDots через клонирование массива. 
        /// </summary>
        private bool[,] _dots;
        public bool[,] GetDots()
        {
            return (bool[,])_dots.Clone(); 
        }
        public void SetDots(bool[,] value)
        {
            _dots = (bool[,])value.Clone();
            Invalidate();
        }
        
        /// <summary>
        /// Цвет фона области матрицы, неактивных точек и фона строки адреса
        /// </summary>
        private Color _fonColor = Color.White;
        [DefaultValue(typeof(Color), "White"), Category("LCD"), Description("")]
        public Color FonColor
        {
            get { return _fonColor; }
            set
            {
                _fonColor = value;
                Invalidate();
            }
        }
       
        /// <summary>
        /// Цвет активной точки
        /// </summary>
        private Color _dotColor = Color.FromArgb(60, 60, 60);
        [DefaultValue(typeof(Color), "60, 60, 60"), Category("LCD"), Description("")]
        public Color DotColor
        {
            get { return _dotColor; }
            set
            {
                _dotColor = value;
                Invalidate();
            }
        }
       
        /// <summary>
        /// Цвет границ точки
        /// </summary>
        private Color _gridColor = Color.FromArgb(160, 160, 160);
        [DefaultValue(typeof(Color), "160, 160, 160"), Category("LCD"), Description("")]
        public Color GridColor
        {
            get { return _gridColor; }
            set
            {
                _gridColor = value;
                Invalidate();
            }
        }
       
        /// <summary>
        /// Разрешение на редактирование
        /// </summary>
        private bool _canEdit = true;
        [DefaultValue(typeof(bool), "true"), Category("LCD"), Description("")]
        public bool CanEdit
        {
            get { return _canEdit; }
            set
            {                
                _canEdit = value;
                charClear.Enabled = charFill.Enabled = charInverse.Enabled = charUp.Enabled = charLeft.Enabled = charRight.Enabled = charDown.Enabled = charUpDown.Enabled = charLeftRight.Enabled = charCut.Enabled = charPaste.Enabled = _canEdit;
            }
        }

        #region propertyAddr
        /// <summary>
        /// Цвет адреса символа. Адрес символа - под этим подразумевается адрес в таблице символов знакогенератора
        /// Цвет самой строки "адрес" - это штатное свойство ForeColor 
        /// </summary>
        private Color _addrColor = Color.Maroon;
        [DefaultValue(typeof(Color), "Maroon"), Category("LCD"), Description("")]
        public Color AddrColor
        {
            get { return _addrColor; }
            set
            {
                _addrColor = value;
                Invalidate();
            }
        }

        /// <summary>
        /// Адрес символа - под этим подразумевается адрес в таблице символов знакогенератора
        /// </summary>
        private Point _address = new Point(0, 0);
        [DefaultValue(typeof(Point), "0;0"), Category("LCD"), Description("")]
        public Point Address
        {
            get { return _address; }
            set
            {
                _address = value;
                Invalidate();
            }
        }
        #endregion propertyAddr
        #endregion
    }
}
